/**************************************************************************

	File: W5100.c

	Copyright (C) 2011-2012 S.Suwa <www.suwa-koubou.jp>

	This file include interface functions
			 to W5100 LAN chip module.

	2012/1/20	recvfrom(), recv(),
				modified increase read pointer. 

***************************************************************************/

/** INCLUDES **************************************************************/
#include <p24FJ64GA002.h>
#include "typedef.h"
#include "pic_232.h"
#include "pic_print.h"
#include "pic_delay.h"
#include "pic_string.h"
#include "W5100.h"

/** DEFINITION ************************************************************/
#ifndef NULL
	#define	NULL	0
#endif

// W5100 Chip Select,  active low
#define	CS_ON()		LATBbits.LATB12=0
#define	CS_OFF()	LATBbits.LATB12=1

// SPI operation code
#define	SPI_READ_OPERATION_CODE		0x0f
#define	SPI_WRITE_OPERATION_CODE	0xf0


/** FUNCTION PROTOTYPE ****************************************************/
// SPI Basic I/O function
void	SpiSend(uchar send_byte);
uchar	SpiSendRec(uchar send_byte);
uchar	SpiRec(void);

/** VARIABLE **************************************************************/
static	uchar	socket_flg[4];
const static	uchar	socket_open_status[] = {
			SOCK_SR_CLOSED,	// mode is CLOSE
			SOCK_SR_INIT,	// mode is TCP
			SOCK_SR_UDP,	// mode is UDP
			SOCK_SR_IPRAW,	// mode is IPRAW
			SOCK_SR_MACRAW,	// mode is MACRAW
			SOCK_SR_PPPOE,	// mode is PPPOE
		};

#ifdef	__USE_FUNCTION

/*==========================================================================
	W5100 Common Register Set operation
==========================================================================*/
/*--------------------------------------------------------------------------
	Set Mode Register
--------------------------------------------------------------------------*/

void	SetMR(uchar mode)
{
	write_byte(MR, mode);
}


/*--------------------------------------------------------------------------
	Set Gateway IP Address Register
--------------------------------------------------------------------------*/

void	SetGWR(uchar ip[])
{
	write_register(4, GWR, ip);
}


/*--------------------------------------------------------------------------
	Set Subnet Mask Register
--------------------------------------------------------------------------*/

void	SetSUBR(uchar ip[])
{
	write_register(4, SUBR, ip);
}


/*--------------------------------------------------------------------------
	Set Source Hardware Address Register
--------------------------------------------------------------------------*/

void	SetSHAR(uchar mac[])
{
	write_register(6, SHAR, mac);
}


/*--------------------------------------------------------------------------
	Set Source IP Address Register
--------------------------------------------------------------------------*/

void	SetSIPR(uchar ip[])
{
	write_register(4, SIPR, ip);
}


/*--------------------------------------------------------------------------
	Set Interupt Mask Register
--------------------------------------------------------------------------*/

void	SetIMR(uchar mask)
{
	write_byte(IMR, mask);
}


/*--------------------------------------------------------------------------
	Set Retry Time-value Register
--------------------------------------------------------------------------*/

void	SetRTR(uint val)
{
	write_word(RTR, val);
}


/*--------------------------------------------------------------------------
	Set Retry Count Register
--------------------------------------------------------------------------*/

void	SetRCR(uchar count)
{
	write_byte(RCR, count);
}


/*--------------------------------------------------------------------------
	Set RX Memory Size Register
--------------------------------------------------------------------------*/

void	SetRMSR(uchar mem_alloc)
{
	write_byte(RMSR, mem_alloc);
}


/*--------------------------------------------------------------------------
	Set TX Memory Size Register
--------------------------------------------------------------------------*/

void	SetTMSR(uchar mem_alloc)
{
	write_byte(TMSR, mem_alloc);
}


/*--------------------------------------------------------------------------
	Set Authentication Type in PPPoE Register
--------------------------------------------------------------------------*/

void	SetPATR(uint auth_type)
{
	write_word(RATR, auth_type);
}


/*--------------------------------------------------------------------------
	Set PPP LCP Request Timer Register
--------------------------------------------------------------------------*/

void	SetPTIMER(uchar val)
{
	write_byte(PTIMER, val);
}


/*--------------------------------------------------------------------------
	Set PPP LCP Magic Number Register
--------------------------------------------------------------------------*/

void	SetPMAGIC(uchar number)
{
	write_byte(PMAGIC, number);
}


/*--------------------------------------------------------------------------
	Set Unreachable IP Address Register
--------------------------------------------------------------------------*/

void	SetUIPR(uchar ip[])
{
	write_register(4, UIPR, ip);
}


/*--------------------------------------------------------------------------
	Set Unreachable Port Register
--------------------------------------------------------------------------*/

void	SetUPORT(uint port)
{
	write_word(UPORT, port);
}


/*==========================================================================
	W5100 Common Register Get operation
==========================================================================*/
/*--------------------------------------------------------------------------
	Get Mode Register
--------------------------------------------------------------------------*/

uchar	GetMR(void)
{
	return read_byte(MR);
}


/*--------------------------------------------------------------------------
	Get Gateway IP Address Register
--------------------------------------------------------------------------*/

void	GetGWR(uchar ip[])
{
	read_register(4, GWR, ip);
}

/*--------------------------------------------------------------------------
	Get Subnet Mask Register
--------------------------------------------------------------------------*/

void	GetSUBR(uchar ip[])
{
	read_register(4, SUBR, ip);
}


/*--------------------------------------------------------------------------
	Get Source Hardware Address Register
--------------------------------------------------------------------------*/

void	GetSHAR(uchar mac[])
{
	read_register(6, SHAR, mac);
}


/*--------------------------------------------------------------------------
	Get Source IP Address Register
--------------------------------------------------------------------------*/

void	GetSIPR(uchar ip[])
{
	read_register(4, SIPR, ip);
}


/*--------------------------------------------------------------------------
	Get Interupt Mask Register
--------------------------------------------------------------------------*/

uchar	GetIMR(void)
{
	return read_byte(IMR);
}


/*--------------------------------------------------------------------------
	Get Retry Time-value Register
--------------------------------------------------------------------------*/

uint	GetRTR(void)
{
	return read_word(RTR);
}


/*--------------------------------------------------------------------------
	Get Retry Count Register
--------------------------------------------------------------------------*/

uchar	GetRCR(void)
{
	return read_byte(RCR);
}


/*--------------------------------------------------------------------------
	Get RX Memory Size Register
--------------------------------------------------------------------------*/

uchar	GetRMSR(void)
{
	return read_byte(RMSR);
}


/*--------------------------------------------------------------------------
	Get TX Memory Size Register
--------------------------------------------------------------------------*/

uchar	GetTMSR(void)
{
	return read_byte(TMSR);
}


/*--------------------------------------------------------------------------
	Get Authentication Type in PPPoE Register
--------------------------------------------------------------------------*/

uint	GetPATR(void)
{
	return	read_word(RATR);
}


/*--------------------------------------------------------------------------
	Get PPP LCP Request Timer Register
--------------------------------------------------------------------------*/

uchar	GetPTIMER(void)
{
	return	read_byte(PTIMER);
}


/*--------------------------------------------------------------------------
	Get PPP LCP Magic Number Register
--------------------------------------------------------------------------*/

uchar	GetPMAGIC(void)
{
	return	read_byte(PMAGIC);
}


/*--------------------------------------------------------------------------
	Get Unreachable IP Address Register
--------------------------------------------------------------------------*/

void	GetUIPR(uchar ip[])
{
	read_register(4, UIPR, ip);
}


/*--------------------------------------------------------------------------
	Get Unreachable Port Register
--------------------------------------------------------------------------*/

uint	GetUPORT(void)
{
	return	read_word(UPORT);
}


#endif	// __USE_FUNCTION

/*==========================================================================
	clear socket flag
===========================================================================*/

void	clear_socket_flg(void)
{
	uchar n;
	for(n = 0; n < 4; n++) socket_flg[n] = 0;
}


/*==========================================================================
	socket()	create socket, handle open
		ip_proto, RAW mode only.
===========================================================================*/

int	socket(uchar mode, uint port, uchar ip_proto)
{
	int		sck_fd;
	uchar	stat;

	// check free socket exists? 
	for(sck_fd = 0; sck_fd < MAX_SOCKET; sck_fd++){
		if(socket_flg[sck_fd] == 0){
			socket_flg[sck_fd] = 1;
			break;
		}
	}
	if(sck_fd >= MAX_SOCKET) return -1;	// no more sockets.

	// check mode parameter
	if((mode & 0x0f) > SOCK_MR_PPPOE) return -1;	// mode error.
	if(((mode & 0x0f) != SOCK_MR_UDP) && (mode & SOCK_MR_MULTI)) return -1; // MULTI is UDP only.
	if(((mode & 0x0f) != SOCK_MR_TCP) && (mode & SOCK_MR_ND)) return -1; // ND is TCP only.

	// set MODE register
	SetSnMR(sck_fd, mode);
	//socket_mode[sck_fd] = mode & 0x0f;	// omitting ND/MULTICAST
	mode &= 0x0f;

	// set PORT, PROTOCOL 
	switch(mode){
	case SOCK_MR_TCP:
	case SOCK_MR_UDP:
		SetSnPORT(sck_fd, port);
		break;
	case SOCK_MR_IPRAW:
		SetSnPROTO(sck_fd, ip_proto);
	}

	// execute socket open
	SetSnCR(sck_fd, SOCK_CR_OPEN);
	// wait command complete.
	while(GetSnCR(sck_fd) != 0);	// 0 value is command complete. 

	// check status
	stat = GetSnSR(sck_fd);
	if(stat != socket_open_status[mode]) return -1;

	// success return
	return sck_fd;
}


/*==========================================================================
	closesocket()	socket handle close
===========================================================================*/

int	closesocket(int sck_fd)
{
	// check asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// release socket
	socket_flg[sck_fd] = 0;

	// execute socket close
	SetSnCR(sck_fd, SOCK_CR_CLOSE);
	while(GetSnCR(sck_fd) != 0);	// 0 value is command complete. 

	// wait for status change to CLOSE
	while(GetSnSR(sck_fd) != SOCK_SR_CLOSED);

	// close success
	return 0;
}

#ifdef	__USE_TCP

/*==========================================================================
	listen()	to server mode
===========================================================================*/

int	listen(int sck_fd)
{
	// check socket asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// LISTEN start from INIT only.
	if(GetSnSR(sck_fd) != SOCK_SR_INIT) return -1;

	// CONNECT command
	SetSnCR(sck_fd, SOCK_CR_LISTEN);
	while(GetSnCR(sck_fd) != 0);	// command complete

	// wait for status change to LISTEN
	while(GetSnSR(sck_fd) != SOCK_SR_LISTEN);

	return 0;	// listen success complete
}


/*==========================================================================
	connect()	connect to remote host (TCP only)
===========================================================================*/

int	connect(int sck_fd, uchar *to_ip, uint to_port)
{
	uchar	status;

	// check socket asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// check parameter
	if(to_ip == NULL || to_port == 0) return -1;
	if( *((ulong *)to_ip) == 0xffffffff ) return -1;

	// CONNECT start from INIT only.
	if(GetSnSR(sck_fd) != SOCK_SR_INIT) return -1;

	// set IP/PORT
	SetSnDIPR(sck_fd, to_ip);
	SetSnDPORT(sck_fd, to_port);

	// CONNECT command
	SetSnCR(sck_fd, SOCK_CR_CONNECT);
	while(GetSnCR(sck_fd) != 0);	// command complete

	// check status
	while((status = GetSnSR(sck_fd)) != SOCK_SR_ESTABLISHED){
		if(status == SOCK_SR_CLOSED) {
			socket_flg[sck_fd] = 0;
			return -1;
		}
	}

	return 0;	// connect success complete
}


/*==========================================================================
	disconnect()	socket disconnect( in ESTABLISH, CLOSE_WAIT )
===========================================================================*/

int	disconnect(int sck_fd)
{
	// check socket asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// DISCONNECT command
	SetSnCR(sck_fd, SOCK_CR_DISCON);
	while(GetSnCR(sck_fd) != 0);	// command complete

	// wait for status change to CLOSE
	while(GetSnSR(sck_fd) != SOCK_SR_CLOSED);

	// release socket
	socket_flg[sck_fd] = 0;

	return 0;
}


/*==========================================================================
	send()	sending data to remote HOST (TCP)
	flag is NONE_BLOCK / BLOCK
===========================================================================*/

int	send(int sck_fd, uchar *buf, uint len, uint flag)
{
	uint	send_size;
	uint	write_ptr;

	// check socket asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// check parameter
	if(buf == NULL || len == 0) return -1;

	// check status
	if(GetSnSR(sck_fd) != SOCK_SR_ESTABLISHED) return 0; // closing or fin close wait.

	// check TX memory free size?
	while((send_size = GetSnTXFSR(sck_fd)) == 0){
		if(flag == NONE_BLOCK) return 0;	// NONE BLOCKING
	}

	// get write pointer
	write_ptr = GetSnTXWR(sck_fd);

	// check write length
	if(send_size > len) send_size = len;

	// data write to memory
	write_memory(sck_fd, write_ptr, buf, send_size);

	// pointer update
	write_ptr += send_size;
	SetSnTXWR(sck_fd, write_ptr);

	// send command
	SetSnCR(sck_fd, SOCK_CR_SEND);
	// wait command complete.
	while(GetSnCR(sck_fd) != 0);	// 0 value is command complete. 

	// wait sending complete
	while(GetSnTXRD(sck_fd) != write_ptr);

	return send_size;
}


/*==========================================================================
	recv()	receiving data from remote terminal (TCP)
	flag is NONE_BLOCK / BLOCK
	return code is received data size.
	if received disconnect@from peer, size was set to Zero, 
===========================================================================*/

int	recv(int sck_fd, uchar *buf, uint len, uint flag)
{
	uint	read_len;
	uint	read_ptr;

	// check asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// check parameter
	if(buf == NULL || len == 0) return -1;

	// check status
	if(GetSnSR(sck_fd) != SOCK_SR_ESTABLISHED) return 0;	// closing or fin close wait.

	// received data exists?
	while((read_len = GetSnRXRSR(sck_fd)) == 0){
		if(flag == NONE_BLOCK) return 0;	// NONE BLOCKING
	}

	// set read length
	//if(read_len > len) read_len = len;

	// get read pointer
	read_ptr = GetSnRXRD(sck_fd);

	// read from RX memory
	read_memory(sck_fd, read_ptr, buf, read_len < len ? read_len: len);

	// update pointer
	read_ptr += read_len;
	SetSnRXRD(sck_fd, read_ptr);

	// update rxbuf pointer
	SetSnCR(sck_fd, SOCK_CR_RECV );
	// wait command complete.
	while(GetSnCR(sck_fd) != 0);	// 0 value is command complete. 

 	return	(read_len < len ? read_len: len);
}

#endif	// __USE_TCP

#ifdef	__USE_UDP

/*==========================================================================
	sendto()	sendinf data to remote terminal (UDP/IPRAW/MACRAW)
	flag is NONE_BLOCK / BLOCK
===========================================================================*/

int	sendto(int sck_fd, uchar *buf, uint len, uint flag, uchar *to_ip, uint to_port )
{
	uint	send_size;
	uint	write_ptr;

	// check socket asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;

	// check parameter
	if(buf == NULL || len == 0 || to_ip == NULL || to_port == 0) return -1;

	// check TX memory free size?
	while((send_size = GetSnTXFSR(sck_fd)) == 0){
		if(flag == NONE_BLOCK) return 0;	// NONE BLOCKING
	}

	// set peer IP/PORT
	SetSnDIPR(sck_fd, to_ip);
	SetSnDPORT(sck_fd, to_port);

	// get write pointer
	write_ptr = GetSnTXWR(sck_fd);

	// check write length
	if(send_size > len) send_size = len;

	// data write to memory
	write_memory(sck_fd, write_ptr, buf, send_size);

	// pointer update
	write_ptr += send_size;
	SetSnTXWR(sck_fd, write_ptr);

	// send command
	SetSnCR(sck_fd, SOCK_CR_SEND);
	// wait command complete.
	while(GetSnCR(sck_fd) != 0);	// 0 value is command complete. 

	// wait sending complete
	while(GetSnTXRD(sck_fd) != write_ptr);

	return send_size;
}


/*==========================================================================
	recvfrom()	receiving data from remote terminal (UDP/IPRAW/MACRAW)
	flag is NONE_BLOCK / BLOCK
===========================================================================*/
#define	UDP_HD_SIZE		8
#define	IPRAW_HD_SIZE	6

int	recvfrom(int sck_fd, uchar *buf, uint len, uint flag, uchar *from_ip, uint *from_port)
{
	uchar	header[UDP_HD_SIZE];	// UDP/IPRAW/MACRAW header
	uint	rec_size;	// received data size
	uint	read_ptr;	// RX memory read pointer
	uint	read_len;	// reading length

	// check socket asign flag
	if(sck_fd < 0 || sck_fd >= MAX_SOCKET ||  socket_flg[sck_fd] != 1) return -1;
	if(buf == NULL || len == 0) return -1;

	// received data exists?
	while((rec_size = GetSnRXRSR(sck_fd)) == 0){
		if(flag == NONE_BLOCK) return 0;	// NONE BLOCKING
	}


	// get read pointer
	read_ptr = GetSnRXRD(sck_fd);


	switch(GetSnSR(sck_fd)){
	case SOCK_SR_UDP:	// UDP

		// reading UDP header
		read_memory(sck_fd, read_ptr, header, UDP_HD_SIZE);
		read_ptr += UDP_HD_SIZE;	// pointer update

		// set from_ip/from_port
		if(from_ip != NULL)	memory_copy(from_ip, header, 4);
		if(from_port != NULL) *from_port = ((uint)header[4] << 8 ) + header[5];

		// get received data length from UDP header
		read_len = ((uint)header[6] << 8) + header[7];

		break;

	case SOCK_SR_IPRAW:	// IPRAW


		// reading IPRAW header
		read_memory(sck_fd, read_ptr, header, IPRAW_HD_SIZE);
		read_ptr += IPRAW_HD_SIZE;	// pointer update

		// set from_ip/from_port
		if(from_ip != NULL)	memory_copy(from_ip, header, 4);

		// get received data length from IPRAW header
		read_len = ((uint)header[4] << 8) + header[5];

		break;

	case SOCK_SR_MACRAW:	// MACRAW


		// reading MACRAW header(data length)
		read_memory(sck_fd, read_ptr, (uchar *)&read_len, sizeof(read_len));
		bg2lt((uchar *)&read_len);	// convert big to little
		read_len -= sizeof(read_len); // data length include header size
		read_ptr += sizeof(read_len); // pointer update

		break;

	default:
		return -1;
	}


	// check data lengh over buffer size
	//if(read_len > len) read_len = len;
	// reading DATA
	// read size limmit to buffer length
	read_memory(sck_fd, read_ptr, buf, read_len < len ? read_len: len);

	// pointer update
	read_ptr += read_len;
	SetSnRXRD(sck_fd, read_ptr);


	// update rxbuf pointer
	SetSnCR(sck_fd, SOCK_CR_RECV );
	// wait command complete.
	//while(GetSnCR(sck_fd) != 0);	// 0 value is command complete. 

 	return	(read_len < len ? read_len: len);
}

#endif	// __USE_UDP

/*==========================================================================
	socket register set function
===========================================================================*/

/*==========================================================================
	socket register get function
===========================================================================*/
/*---------------------------------------------------------------------------
	get sending free memory  size
----------------------------------------------------------------------------*/

uint	GetSnTXFSR(int sck_fd)
{
	uint	old, new;

	old = read_word(Sn_TX_FSR + sck_fd * SOCKET_REGISTER_SIZE);
	while(old != (new = read_word(Sn_TX_FSR + sck_fd * SOCKET_REGISTER_SIZE))){
		old = new;
	}
	return new;
}


/*---------------------------------------------------------------------------
	get received data size
----------------------------------------------------------------------------*/

uint	GetSnRXRSR(int sck_fd)
{
	uint	old, new;

	old = read_word(Sn_RX_RSR + sck_fd * SOCKET_REGISTER_SIZE);
	while(old != (new = read_word(Sn_RX_RSR + sck_fd * SOCKET_REGISTER_SIZE))){
		old = new;
	}
	return new;
}


/*==========================================================================
	socket TX/RX memory read write function
===========================================================================*/
/*---------------------------------------------------------------------------
	write data to socket TX memory
----------------------------------------------------------------------------*/

void	write_memory(int sck_fd, uint write_ptr, uchar *buf, uint len)
{
	uint	memory_addr, offset;
	uint	upper_size, left_size;

	// calculate offset address 
	offset = write_ptr & (TX_MEMORY_SIZE -1);

	// calculate physical memory start address
	memory_addr = TX_MEMORY_BASE_ADDR + sck_fd * TX_MEMORY_SIZE + offset;

	// if overflow socket TX memory ?
	if(offset + len > TX_MEMORY_SIZE){

		// copy upper_size bytes
		upper_size = TX_MEMORY_SIZE - offset;
		write_register(upper_size, memory_addr, buf);
		buf += upper_size;

		// copy left size bytes
		left_size = len - upper_size;
		write_register(left_size, TX_MEMORY_BASE_ADDR + sck_fd * TX_MEMORY_SIZE, buf);

	}else{

		// copy len size bytes
		write_register(len, memory_addr, buf);
	}
}


/*---------------------------------------------------------------------------
	read data from socket RX memory
----------------------------------------------------------------------------*/

void	read_memory(int sck_fd, uint read_ptr, uchar *buf, uint len)
{
	uint	memory_addr, offset;
	uint	upper_size, left_size;

	// calculate offset address 
	offset = read_ptr & (RX_MEMORY_SIZE -1);

	// calculate physical memory start address
	memory_addr = RX_MEMORY_BASE_ADDR + sck_fd * RX_MEMORY_SIZE + offset;

	// if overflow socket RX memory ?
	if(offset + len > RX_MEMORY_SIZE){

		// copy upper_size bytes
		upper_size = RX_MEMORY_SIZE - offset;
		read_register(upper_size, memory_addr, buf);
		buf += upper_size;

		// copy left size bytes
		left_size = len - upper_size;
		read_register(left_size, RX_MEMORY_BASE_ADDR + sck_fd * RX_MEMORY_SIZE, buf);

	}else{

		// copy len size bytes
		read_register(len, memory_addr, buf);
	}
}


/*==========================================================================
	W5100 Register/Memory, read, write basic function
==========================================================================*/
/*--------------------------------------------------------------------------
	write to register
--------------------------------------------------------------------------*/

void	write_register(int octet, uint reg_addr, uchar *data)
{
	int	n;

	for(n = 0; n < octet; n++){
		write_byte(reg_addr++, *data++);
	}
}


/*--------------------------------------------------------------------------
	write_word()	write word operation	
--------------------------------------------------------------------------*/

void	write_word(uint reg_addr, uint data)
{
	write_byte(reg_addr++, (data >> 8) & 0xff);
	write_byte(reg_addr, data & 0xff);
}


/*--------------------------------------------------------------------------
	write_byte()	write byte operation
--------------------------------------------------------------------------*/

void	write_byte(uint addr, uchar data)
{
	CS_ON();
	SpiSend(SPI_WRITE_OPERATION_CODE);
	SpiSend((addr & 0xff00)>>8);
	SpiSend(addr & 0xff);
	SpiSend(data);
	CS_OFF();
}


/*--------------------------------------------------------------------------
	read from register
--------------------------------------------------------------------------*/

void	read_register(int octet, uint reg_addr, uchar *data)
{
	int	n;

	for(n = 0; n < octet; n++){
		*data++ = read_byte(reg_addr++);
	}
}


/*--------------------------------------------------------------------------
	read_word()	read word operation
--------------------------------------------------------------------------*/

uint	read_word(uint reg_addr)
{
	uint	data;

	data = (uint)read_byte(reg_addr++);
	data <<= 8;
	data += (uint)read_byte(reg_addr);

	return data;
}


/*--------------------------------------------------------------------------
	read_byte()	read byte operation
--------------------------------------------------------------------------*/

uchar	read_byte(uint addr)
{
	uchar	data;

	CS_ON();
	SpiSend(SPI_READ_OPERATION_CODE);
	SpiSend((addr & 0xff00)>>8);
	SpiSend(addr & 0xff);
	data = SpiRec();
	CS_OFF();
	return data;
}


/*==========================================================================
	other utility
==========================================================================*/
/*--------------------------------------------------------------------------
	word value convert to little endian from big endian. 
--------------------------------------------------------------------------*/

void	bg2lt(uchar val[])
{
	uchar	d;
	d = val[0];
	val[0] = val[1];
	val[1] = d;
}

/*----------------------------------------------------------------
	htons()
----------------------------------------------------------------*/

uint	htons(uint val)
{
	return ((val << 8)& 0xff00) | ((val >> 8) & 0xff);
}


/*-------------------------------------------------------------------------
	htonl()
-------------------------------------------------------------------------*/

ulong	htonl(ulong val)
{
	uint	vu, vl;

	vu = val & 0xffff;
	vl = (val >> 16) & 0xffff;
	vu = htons(vu);
	vl = htons(vl);

	return ((ulong)vu << 16) | vl;
}


/*--------------------------------------------------------------------------
	ntohs()
--------------------------------------------------------------------------*/

uint	ntohs(uint val)
{
	return htons(val);
}


/*---------------------------------------------------------------------------
	ntohl()
---------------------------------------------------------------------------*/

ulong	ntohl(ulong val)
{
	return htonl(val);
}

/*----------------------------------------------------------------------
	set ip/mac
		uchar ip[4], mac[6];
		set_ip(ip, "192.168.1.20")
		set_mac(mac, "00.01.12.23.34.45");
----------------------------------------------------------------------*/

void	set_ip(uchar ip[], char *ip_str)
{
	int		n;

	ip[0] = atoi(ip_str);
	for(n = 1; n < 4; n++){
		ip_str = strchr(ip_str, '.');
		if(ip_str == NULL) return;
		ip_str++;	// skip '.'
		ip[n] = atoi(ip_str);
	}

}


void	set_mac(uchar mac[], char *mac_str)
{
	int		n;

	mac[0] = htoi(mac_str);
	for(n = 1; n < 6; n++){
		mac_str = strchr(mac_str, '.');
		if(mac_str == NULL) return;
		mac_str++;	// skip '.'
		mac[n] = htoi(mac_str);
	}

}


/*----------------------------------------------------------------------
	print ip/mac
----------------------------------------------------------------------*/

char *print_ip(char *strbuf, uchar ip[])
{
	int		n;
	char	*s;

	s = strbuf;
	uint_to_decimal(s, (uint)ip[0]); s = strchr(s, 0);
	for(n = 1; n < 4; n++){
		*s = '.'; s++;
		uint_to_decimal(s, (uint)ip[n]); s = strchr(s, 0);
	}
	return strbuf;
}

char *print_mac(char *strbuf, uchar mac[])
{
	int		n;
	char	*s;

	s = strbuf;
	int_to_2hex(s, (uint)mac[0]); s = strchr(s, 0);
	for(n = 1; n < 6; n++){
		*s = '.'; s++;
		int_to_2hex(s, (uint)mac[n]); s = strchr(s, 0);
	}
	return strbuf;
}


/*==========================================================================
	Spi basic function
==========================================================================*/
/*--------------------------------------------------------------------------
	SpiSend()	received data ignore
--------------------------------------------------------------------------*/

void	SpiSend(unsigned char send_byte)
{
	unsigned char dummy;

	// wait until TXBUF empty
	while(SPI1STATbits.SPITBF);

	// write to TXBUF
	SPI1BUF = send_byte;

	// wait until data send out
	//  (wait for data received complete)
	while(!SPI1STATbits.SPIRBF);

	dummy = SPI1BUF;	// clear receive buffer 
}


/*--------------------------------------------------------------------------
	SpiSendRec()	data send and data receive 
--------------------------------------------------------------------------*/

unsigned char	SpiSendRec(unsigned char send_byte)
{
	// wait until TXBUF empty
	while(SPI1STATbits.SPITBF);

	// write to TXBUF
	SPI1BUF = send_byte;

	// wait until data send out
	//  (wait for data received complete)
	while(!SPI1STATbits.SPIRBF);

	return SPI1BUF;
}


/*--------------------------------------------------------------------------
	SpiRec()	data receive( dummy send ) 
--------------------------------------------------------------------------*/

unsigned char	SpiRec(void)
{
	// wait until TXBUF empty
	while(SPI1STATbits.SPITBF);

	// write to dummy data
	SPI1BUF = 0xff;

	// wait until data send out
	//  (wait for data received complete)
	while(!SPI1STATbits.SPIRBF);

	return SPI1BUF;
}


/*** end W5100.c ************************************************************/

