/***************************************************************************

	File: key.c

	Copyright (C) 2012 S.Suwa <www.suwa-koubou.jp>

***************************************************************************/

/** INCLUDES **************************************************************/
#include <p24FJ64GA002.h>
#include "typedef.h"
#include "pic_232.h"
#include "pic_print.h"
#include "pic_delay.h"
#include "pic_string.h"
#include "W5100.h"

#include "key.h"

/** DEFINITION ************************************************************/ 
/**
  keyunit key scan input level
              register-value ratio  ratio*128  judge-level range
    button 1       0/1000     0		    0		0 -   8      9
    button 2     150/1150	  0.13	   17       9 -  25     17
    button 3     330/1330	  0.248	   32      26 -  40     15
    button 4     600/1600     0.375	   48      41 -  55     15
    button 5     990/1990     0.498	   64      56 -  70     15
    button 6    1500/2500     0.6      77      71 -  85     15
    button 7    2410/3410     0.71     91      86 - 100     15
    none                      1.0     128     101 - 127     27						

  keyunit button layout
    ---------------------
    button  k1   k2   k3
    ---------------------
       1    ON   <-   ->     left_arrow and right_arrow are
       2    OFF  ENT  CLR	combination key. when left/right arrow
       3    1    2    3     are pressed same time, create mode change key.   
       4    4    5    6
       5    7    8    9
       6    .    0    #
    ---------------------
***/

// IPphone does not use button7. 
const char	key_scan[128] = {
	1, 1, 1, 1, 1, 1, 1, 1,
	1, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 3, 3, 3, 3, 3, 3,
	3, 3, 3, 3, 3, 3, 3, 3,
	3, 4, 4, 4, 4, 4, 4, 4,
	4, 4, 4, 4, 4, 4, 4, 4, 
	5, 5, 5, 5, 5, 5, 5, 5,
	5, 5, 5, 5, 5, 5, 5, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	};

const int	key_code[3][7] = {
		{ KEY_NONE, KEY_ON_HOOK, KEY_OFF_HOOK, '1', '4', '7', '*', },
		{ KEY_NONE, KEY_LEFT,	 KEY_ENTER,    '2', '5', '8', '0', },
		{ KEY_NONE, KEY_RIGHT,   KEY_CLEAR,	   '3', '6', '9', '#', },
	};


static volatile int	key_raw[3];
extern volatile int	mic_raw;	// mic_raw definition in sound.c	

volatile int	key_int_count;
volatile int	key_match_count;
static volatile int nkey1, nkey2, nkey3;
static volatile int pkey1, pkey2, pkey3;

/*----------------------------------------------------------------------
	AD1 conversion done, 25usec interval interrupt
----------------------------------------------------------------------*/

void __attribute__ ((__interrupt__, no_auto_psv)) _ADC1Interrupt(void)
{
	IFS0bits.AD1IF = 0;		// clear interrupt flag
	nkey1 = key_scan[ADC1BUF0 >> 3]; // 10bits to 7bits
	nkey2 = key_scan[ADC1BUF1 >> 3]; //  1023 -> 127
	nkey3 = key_scan[ADC1BUF2 >> 3];
	mic_raw	= ADC1BUF3 -512;	// convert unsigned to signed
	
	if(--key_int_count) return;
	key_int_count = INT_TIMES;

	if(pkey1 == nkey1 && pkey2 == nkey2 && pkey3 == nkey3){
		if(--key_match_count == 0){
			key_raw[0] = key_code[0][nkey1];
			key_raw[1] = key_code[1][nkey2];
			key_raw[2] = key_code[2][nkey3];
			key_match_count = MATCH_TIMES;
		}
	}else{
		pkey1 = nkey1;
		pkey2 = nkey2;
		pkey3 = nkey3;
		key_match_count = MATCH_TIMES;
	}
}


/*--------------------------------------------------------------
	getkey()  This function is none blocking
--------------------------------------------------------------*/

int		getkey(void)
{
	static  int prev_key, now_key;
			int	now[3];

	// get key
	asm("DISI #100");
	now[0] = key_raw[0];
	now[1] = key_raw[1];
	now[2] = key_raw[2];
	asm("DISI #0");

	// key check
	if(now[1] == KEY_LEFT && now[2] == KEY_RIGHT){
		now_key = KEY_MODE;

	}else if(now[0] != KEY_NONE && now[1] == KEY_NONE && now[2] == KEY_NONE){
		now_key = now[0];

	}else if(now[0] == KEY_NONE && now[1] != KEY_NONE && now[2] == KEY_NONE){
		now_key = now[1];

	}else if(now[0] == KEY_NONE && now[1] == KEY_NONE && now[2] != KEY_NONE){
		now_key = now[2];

	}else{
		prev_key = now_key = KEY_NONE;
	}

	// changed key ?
	if(now_key != prev_key){
		prev_key = now_key;
		return now_key;
	}else{
		return -1;	// same key press or no key press 
	}
}


/*** end of key.c *********************************************************/
