/************************************************************************************

	File: lcd.c

	Copyright (C) 2012 S.Suwa <www.suwa-koubou.jp>


	Lcd Display module Control

 icon address  asign bit
	0x00, 	0b10000, // Antenna
	0x02, 	0b10000, // Telephone
	0x04, 	0b10000, // Speaker Sound
	0x06, 	0b10000, // Right Arrow
	0x07, 	0b10000, // Up Arrow
	0x07, 	0b01000, // Down Arrow
	0x09, 	0b10000, // Key Lock
	0x0B, 	0b10000, // Don't 
	0x0D, 	0b10000, // Battery (low level)
	0x0D, 	0b01000, // Battery (middle level)
	0x0D, 	0b00100, // Battery (high level)
	0x0D, 	0b00010, // Battery
	0x0F, 	0b10000, // Flower

************************************************************************************/

/** INCLUDES ***********************************************************************/
#include <p24FJ64GA002.h>
#include "typedef.h"
#include "pic_232.h"
#include "pic_print.h"
#include "pic_delay.h"
#include "lcd.h"

/** DEFINITION *********************************************************************/
#define	LCD_DEVICE	0b01111100
#define	MW	0	// I2C master write
#define	MR	1	// I2C master read


/*-----------------------------------------------------------------------------------
	lcd initialize
-----------------------------------------------------------------------------------*/

void	lcd_initialize(void)
{
	unsigned char contrast;

	//*** lcd initialize **********************************************
	// set contrast control value
	contrast = 0b111100;

	// power on after 40msec
	delay_ms(40);

	// function 8bit,N=1,5*8dot
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

	// select extension instruction
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);

	// internal OSC frequency adjustment, by default value(0x04)
	lcd_cmdwrite(cmd_OSC_ADJ | 0x04);	

	// set contrast low
	lcd_cmdwrite(cmd_CONTRAST_L | (contrast & 0b1111));

	// set contrast high and icon turned on
	lcd_cmdwrite(cmd_ICON_CONTRAST_H | prm_ICON_ON | ((contrast >> 4) & 0b11));

	// fllower control
	lcd_cmdwrite(cmd_FOLLOWER);

	// wait for LCD device stable
	delay_ms(200);

	// function 8bit,N=1,5*8dot
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

	// Display ON
	lcd_cmdwrite(cmd_DISPLAY | prm_DISPLAY_ON | prm_CURSOR_OFF | prm_BLINK_OFF); // Display On

	// Clear Display
	lcd_cmdwrite(cmd_CLEAR);	// Clear Display
	delay_ms(2);		// wait for clear display

	// Battery ICON on
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR +0x0d); // set ICON address
	lcd_datawrite(0b11110);	// set ICON bit
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

}

/*-----------------------------------------------------------------------------------
	telephone icon on/off
-----------------------------------------------------------------------------------*/

void	icon_telephone(int sw)
{
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR +2); // set ICON address
	if(sw == ON){
		lcd_datawrite(0b10000);	// set ICON bit
	}else{
		lcd_datawrite(0b00000);	// clear ICON bit
	}
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

}


/*-----------------------------------------------------------------------------------
	ring bell icon on/off
-----------------------------------------------------------------------------------*/

void	icon_ring_bell(int sw)
{
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR +4); // set ICON address
	if(sw == ON){
		lcd_datawrite(0b10000);	// set ICON bit
	}else{
		lcd_datawrite(0b00000);	// clear ICON bit
	}
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);
}


/*-----------------------------------------------------------------------------------
	command write
-----------------------------------------------------------------------------------*/

void	lcd_cmdwrite(unsigned char cmd)
{
	i2cOpenStart();
	i2cSend(LCD_DEVICE +MW);
	i2cSend(0b00000000);	// RS=0, select instruction register
	i2cSend(cmd);
	i2cStopClose();
	delay_ms(1);
}


/*-----------------------------------------------------------------------------------
	data write
-----------------------------------------------------------------------------------*/

void	lcd_datawrite(char data)
{
	i2cOpenStart();
	i2cSend(LCD_DEVICE +MW);
	i2cSend(0b01000000);	// RS=1, select data register
	i2cSend((uchar)data);
	i2cStopClose();
	delay_ms(1);
}


/*-----------------------------------------------------------------------------------
	string write
-----------------------------------------------------------------------------------*/

void	lcd_strwrite(char *str)
{
	while(*str){
		lcd_datawrite(*str++);
	}
}


/*==========================================================================
	I2C1 Handling module
==========================================================================*/
#define	TIME_OUT	0xffff
static	unsigned short tw;

/*--------------------------------------------------------------------------
	I2C1 master Open and Start
--------------------------------------------------------------------------*/

void	i2cOpenStart(void)
{
	// enabled I2C1
	I2C1CONbits.I2CEN = 1;
	delay_1us();

	// START 
	I2C1CONbits.SEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1CONbits.SEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C1 master ReStart
--------------------------------------------------------------------------*/

void	i2cReStart(void)
{
	// RESTART 
	I2C1CONbits.RSEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1CONbits.RSEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C1 master Stop and Close
--------------------------------------------------------------------------*/

void	i2cStopClose(void)
{
	// STOP 
	I2C1CONbits.PEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1CONbits.PEN)break;
	}
	delay_1us();

	// disabled I2C1
	I2C1CONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
	I2C1 master data send
--------------------------------------------------------------------------*/

int	i2cSend(unsigned char data)
{
	I2C1TRN = data;

    if(I2C1STATbits.IWCOL){	// If write collision occurs,return -2
		return -2;
    }

	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1STATbits.TRSTAT) break;  // wait until write cycle is complete 
	}

	if (I2C1STATbits.ACKSTAT){ 	// test for ACK condition received
		return -1;	// NACK received
	}

	return 0;	// complete with no error.
}


/*--------------------------------------------------------------------------
	I2C1	master data receive
--------------------------------------------------------------------------*/

unsigned char	i2cRec(void)
{
	// enabled Receive
	I2C1CONbits.RCEN = 1;

	// wait until receive complete and data stored to I2C1RCV register
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1CONbits.RCEN && I2C1STATbits.RBF) break;
	}

	// clear overflow bit
	I2C1STATbits.I2COV = 0;
    
	return I2C1RCV;
}


/*--------------------------------------------------------------------------
	I2C1 master Ack send
--------------------------------------------------------------------------*/

void	i2cAckSend(void)
{
	// Ack Send
	I2C1CONbits.ACKDT = 0;
	I2C1CONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1CONbits.ACKEN)break;
	}
}


/*--------------------------------------------------------------------------
	I2C1 master NotAck send
--------------------------------------------------------------------------*/

void	i2cNAckSend(void)
{
	// NAck Send
	I2C1CONbits.ACKDT = 1;
	I2C1CONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C1CONbits.ACKEN) break;
	}
}


/*** end of lcd.c **************************************************************/
