/*********************************************************************

	File: pic_timer.c

	Copyright (C) 2011-2012 S.Suwa <www.suwa-koubou.jp>

	This file is application timer module.
	 Using timer T2 100msec interval interrupt.

**********************************************************************/

#include	<p24FJ64GA002.h>

/*** definition **************************************************/
#define	MAX_TIMER	5
#define	NULL		(void *)0

typedef struct	timer_t {
	unsigned int flag;	// 1=use, 0=free
	unsigned int count;	// down count timer
	unsigned int save;	// saved counter initial value
	void (*call_back)();	// if timer_count is zero, this function called back. 
} TIMER_TABLE;

/*** private variable **************************************************/
static	TIMER_TABLE	timer_table[MAX_TIMER];

/*** global variable **************************************************/
unsigned int timer_clock;	// 0.1sec increment counter, during power on.

/*-----------------------------------------------------------------------
	initilaize timer modelue
-----------------------------------------------------------------------*/

void timer_initialize(void)
{
	int	tmr_id;

	// clear timer table
	for(tmr_id = 0; tmr_id < MAX_TIMER; tmr_id++) {
		timer_table[tmr_id].flag = 0;
		timer_table[tmr_id].count = 0;
		timer_table[tmr_id].call_back = NULL;
	}


	//--- setting interrupt ---------------------------------------
	INTCON1 = 0;		// NSTDIS=0 nesting ok
	IEC0bits.T2IE = 1;	// enable T2 interrupt
	IFS0bits.T2IF = 0;	// clear interrupt flag

	//--- set interval@100ms timer --------------------------------
	T2CON = 0x0020;	// use timer(T2), set prescale 1:64
	TMR2 = 0;
	PR2 = 24999;		// 100msec (25000 * 64 = 1600000 = 1.6M, 16M/1.6M=10Hz,100msec)
	T2CONbits.TON = 1;	// start T2

}


/*----------------------------------------------------------------------
	Timer T2 interrupt
	Timer T2 100msec interval timer.
----------------------------------------------------------------------*/

void __attribute__ ((__interrupt__, no_auto_psv)) _T2Interrupt(void)
{
	int	tmr_id;

	IFS0bits.T2IF = 0;		// clear interrupt flag

	timer_clock++;

	// timer count down
	for(tmr_id = 0; tmr_id < MAX_TIMER; tmr_id++){
		if(timer_table[tmr_id].flag){
			if(timer_table[tmr_id].count){
				if(--timer_table[tmr_id].count == 0){
					if(timer_table[tmr_id].call_back != NULL){
						// interval down count recovery set
						timer_table[tmr_id].count = timer_table[tmr_id].save;
						// call function
						(*(timer_table[tmr_id].call_back))();
					}
				}
			}
		}
	}

}


/*----------------------------------------------------------------------
	timer_entry()
----------------------------------------------------------------------*/

int	timer_entry(unsigned int count, void (*call_back)(void))
{
	int tmr_id;

	for(tmr_id = 0; tmr_id < MAX_TIMER; tmr_id++){
		if(timer_table[tmr_id].flag == 0){ 
			asm("DISI #100");
			timer_table[tmr_id].flag = 1;
			timer_table[tmr_id].count = count;
			timer_table[tmr_id].save = count;
			timer_table[tmr_id].call_back = call_back;
			asm("DISI #0");
			return tmr_id;
		}
	}
	return -1;
}


/*----------------------------------------------------------------------
	timer_update()
----------------------------------------------------------------------*/

void timer_update(int tmr_id, unsigned int count)
{
	if(tmr_id >= 0 && tmr_id < MAX_TIMER){
		asm("DISI #100");
		timer_table[tmr_id].count = count;
		asm("DISI #0");
	}
}


/*----------------------------------------------------------------------
	timer_delete()
----------------------------------------------------------------------*/

void timer_delete(int tmr_id)
{
	if(tmr_id >= 0 && tmr_id < MAX_TIMER){
		asm("DISI #100");
		timer_table[tmr_id].flag = 0;
		asm("DISI #0");
	}
}


/*----------------------------------------------------------------------
	timer_check()
----------------------------------------------------------------------*/

int timer_check(int tmr_id)
{
	if(tmr_id >= 0 && tmr_id < MAX_TIMER){
		return timer_table[tmr_id].count;
	}
	return 0;
}


/*----------------------------------------------------------------------
	timer_Up()
----------------------------------------------------------------------*/

int timer_Up(int tmr_id)
{
	if(tmr_id >= 0 && tmr_id < MAX_TIMER){
		return timer_table[tmr_id].count == 0 ? 1: 0;
	}
	return 0;
}


/*** end of pic_timer.c ***********************************************/

